# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import clize

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.change_batchsize import ChangeBatchSize
from qonnx.transformation.extract_conv_bias import ExtractBiasFromConv
from qonnx.transformation.fold_constants import FoldConstants
from qonnx.transformation.general import (
    GiveReadableTensorNames,
    GiveUniqueNodeNames,
    GiveUniqueParameterTensors,
    RemoveStaticGraphInputs,
    RemoveUnusedTensors,
)
from qonnx.transformation.infer_shapes import InferShapes
from qonnx.transformation.quant_constant_folding import FoldTransposeIntoQuantInit


def cleanup_model(model, preserve_qnt_ops=True, override_inpsize=None, extract_conv_bias=False):
    """Execute the transformations for the cleanup function on a model level.
    This allows the reuse of the cleanup transformations, without needing to read/write the model from/to disk.

    :param model: A raw QONNX model from as example Brevitas.
    :return model_clean: The cleaned model
    """

    # temporary fix for QONNX op domains
    qonnx_domain_ops = ["Quant", "Trunc", "BipolarQuant"]
    for q_op_type in qonnx_domain_ops:
        qnt_nodes = model.get_nodes_by_op_type(q_op_type)
        for qnt_node in qnt_nodes:
            qnt_node.domain = "qonnx.custom_op.general"
    if preserve_qnt_ops:
        preserve_qnt_optypes = ["Quant", "BipolarQuant", "QuantizeLinear", "DequantizeLinear"]
    else:
        preserve_qnt_optypes = []

    if override_inpsize is not None:
        if type(override_inpsize) is str:
            override_inpsize = eval(override_inpsize)
        if type(override_inpsize) is int:
            override_batchsize = override_inpsize
            model = model.transform(ChangeBatchSize(override_batchsize))
        elif type(override_inpsize) is tuple:
            override_batchsize = override_inpsize[0]
            model = model.transform(ChangeBatchSize(override_batchsize))
            iname = model.graph.input[0].name
            model.set_tensor_shape(iname, override_inpsize)

    cleanup_transformations = [
        InferShapes(),
        GiveUniqueParameterTensors(),
        FoldConstants(exclude_op_types=preserve_qnt_optypes),
        FoldTransposeIntoQuantInit(),
        RemoveUnusedTensors(),
        RemoveStaticGraphInputs(),
        GiveUniqueNodeNames(),
        GiveReadableTensorNames(),
    ]
    for t in cleanup_transformations:
        model = model.transform(t)

    if extract_conv_bias:
        model = model.transform(ExtractBiasFromConv())
        model = model.transform(InferShapes())
        model = model.transform(GiveUniqueNodeNames())
        model = model.transform(GiveReadableTensorNames())

    return model


def cleanup(in_file, *, out_file=None, preserve_qnt_ops=True, override_inpsize: str = None, extract_conv_bias=False):
    """Execute a set of graph transformations to clean-up the given ONNX file.

    :param in_file: Filename for the input ONNX model
    :param preserve_qnt_ops: Preserve weight quantization operators
    :param out_file: If set, filename for the output ONNX model. Set to in_file with _clean
        suffix otherwise.
    :param override_inpsize: If specified, override the input size (e.g. "(1,3,224,224)" to set all or
        just 1 to set batchsize to 1) for the ONNX graph
    :param extract_conv_bias: If specified, separate Conv bias into its own Add node
    """

    model = ModelWrapper(in_file)
    model = cleanup_model(
        model, preserve_qnt_ops=preserve_qnt_ops, override_inpsize=override_inpsize, extract_conv_bias=extract_conv_bias
    )
    if out_file is None:
        out_file = in_file.replace(".onnx", "_clean.onnx")
    model.save(out_file)


def main():
    clize.run(cleanup)


if __name__ == "__main__":
    main()



# Copyright (c) 2021 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distributionode.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permissionode.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np

from qonnx.util.basic import get_by_name


def get_node_tensor_dtypes(model, node):
    # input tensor (input 0)
    i_name = node.input[0]
    i_dtype = model.get_tensor_datatype(i_name)
    # weight tensor (input 1)
    w_name = node.input[1]
    w_dtype = model.get_tensor_datatype(w_name)
    # output tensor (input 0)
    o_name = node.output[0]
    o_dtype = model.get_tensor_datatype(o_name)
    return (i_dtype, w_dtype, o_dtype)


def get_node_tensor_shapes(model, node):
    # input tensor (input 0)
    i_name = node.input[0]
    i_shape = model.get_tensor_shape(i_name)
    assert i_shape is not None, "Input has undefined shape: " + str(node)
    # weight tensor (input 1)
    w_name = node.input[1]
    w_shape = model.get_tensor_shape(w_name)
    assert w_shape is not None, "Weight has undefined shape: " + str(node)
    # output tensor (output 0)
    o_name = node.output[0]
    o_shape = model.get_tensor_shape(o_name)
    assert o_shape is not None, "Output has undefined shape: " + str(node)
    return (i_shape, w_shape, o_shape)


def get_node_weight_density(model, w_name):
    w_tensor = model.get_initializer(w_name)
    if w_tensor is None:
        return 1.0
    w_total = np.prod(w_tensor.shape)
    w_density = np.count_nonzero(w_tensor) / w_total
    return w_density


def aggregate_dict_keys(res_dict):
    total_dict = {}
    for layer in res_dict:
        layer_res_dict = res_dict[layer]
        for r_type in layer_res_dict.keys():
            if "efficiency" in r_type:
                continue
            r_amount = layer_res_dict[r_type]
            r_amount = float(r_amount)
            if r_type in total_dict.keys():
                total_dict[r_type] += r_amount
            else:
                total_dict[r_type] = r_amount
    return total_dict


def inference_cost_conv(model, node, discount_sparsity):
    # extract info about the conv kernel attributes
    k = get_by_name(node.attribute, "kernel_shape").ints
    k_prod = np.prod(k)
    group = get_by_name(node.attribute, "group")
    if group is None:
        group = 1
    else:
        group = group.i
    # extract info from tensor shapes and datatypes
    (i_dtype, w_dtype, o_dtype) = get_node_tensor_dtypes(model, node)
    (i_shape, w_shape, o_shape) = get_node_tensor_shapes(model, node)
    bsize = i_shape[0]
    ifm_ch = i_shape[1]
    ofm_ch = o_shape[1]
    assert ofm_ch == w_shape[0], "Mismatch in output channels"
    assert ofm_ch % group == 0, "Invalid group setting: " + str(node)
    ofm_pix_total = np.prod(o_shape[2:])
    n_macs = bsize * (ofm_ch // group) * ifm_ch * k_prod * ofm_pix_total
    w_mem = np.prod(w_shape)
    o_mem = np.prod(o_shape)
    if discount_sparsity:
        wname = node.input[1]
        density = get_node_weight_density(model, wname)
        n_macs *= density
        w_mem *= density
    idt_name = i_dtype.name
    wdt_name = w_dtype.name
    odt_name = o_dtype.name
    mac_op_type_str = "op_mac_%s_%s" % (idt_name, wdt_name)
    w_mem_type_str = "mem_w_%s" % (wdt_name)
    o_mem_type_str = "mem_o_%s" % (odt_name)
    # keep in floats to remain compatible with json serialization
    n_macs, w_mem, o_mem = float(n_macs), float(w_mem), float(o_mem)
    ret = {mac_op_type_str: n_macs, w_mem_type_str: w_mem, o_mem_type_str: o_mem}
    return ret


def inference_cost_matmul(model, node, discount_sparsity):
    # extract info from tensor shapes and datatypes
    (i_dtype, w_dtype, o_dtype) = get_node_tensor_dtypes(model, node)
    (i_shape, w_shape, o_shape) = get_node_tensor_shapes(model, node)
    if node.op_type == "Gemm":
        assert len(i_shape) == 2 and len(w_shape) == 2
        tA = get_by_name(node.attribute, "transA")
        tB = get_by_name(node.attribute, "transB")
        if tA is not None and tA.i == 1:
            i_shape = i_shape[::-1]
        if tB is not None and tB.i == 1:
            w_shape = w_shape[::-1]
    # exclude common dim (last axis) from one side to avoid duplication
    n_macs = i_shape[-1] * np.prod(o_shape)
    # deal with both dyn,param and dyn,dyn cases for weight memory
    inp0_is_const = model.get_initializer(node.input[0]) is not None
    inp1_is_const = model.get_initializer(node.input[1]) is not None
    if inp0_is_const and (not inp1_is_const):
        # inp 0 is static
        w_mem = np.prod(i_shape)
        wname = node.input[0]
    elif (not inp0_is_const) and inp1_is_const:
        # inp 1 is static
        w_mem = np.prod(w_shape)
        wname = node.input[1]
    elif (not inp0_is_const) and (not inp1_is_const):
        # both inputs dynamic
        w_mem = 0
        wname = None
    if discount_sparsity and wname is not None:
        density = get_node_weight_density(model, wname)
        n_macs *= density
        w_mem *= density
    o_mem = np.prod(o_shape)
    idt_name = i_dtype.name
    wdt_name = w_dtype.name
    odt_name = o_dtype.name
    mac_op_type_str = "op_mac_%s_%s" % (idt_name, wdt_name)
    w_mem_type_str = "mem_w_%s" % (wdt_name)
    o_mem_type_str = "mem_o_%s" % (odt_name)
    # keep in floats to remain compatible with json serialization
    n_macs, w_mem, o_mem = float(n_macs), float(w_mem), float(o_mem)
    ret = {mac_op_type_str: n_macs, w_mem_type_str: w_mem, o_mem_type_str: o_mem}
    return ret


def inference_cost_upsample(model, node, discount_sparsity):
    # extract info about the upsampling kernel attributes
    mode = get_by_name(node.attribute, "mode").s.decode("utf-8")
    scales_tensor = node.input[1]
    scales_initializer = model.get_initializer(scales_tensor)

    # extract info from tensor shapes and datatypes
    (i_dtype, scale_dtype, o_dtype) = get_node_tensor_dtypes(model, node)
    (i_shape, scale_shape, o_shape) = get_node_tensor_shapes(model, node)
    bsize = i_shape[0]
    ifm_ch = i_shape[1]
    ofm_pix_total = np.prod(o_shape[2:])

    # MAC calculation
    if mode == "nearest":
        # No calculation involved, since data is just copied over multiple times
        n_macs = 0
    elif mode == "linear":
        # Data gets linearly interpolated in each dimension
        # Two MACs per dimension and output pixel assumed
        n_dim_scaling = np.sum(scales_initializer > 1)
        n_macs = 2 * n_dim_scaling * ofm_pix_total * ifm_ch * bsize
    else:
        raise ValueError(f"Upsampling mode {mode} not supported for estimation.")

    # Mem calculation
    o_mem = np.prod(o_shape)
    idt_name = i_dtype.name
    odt_name = o_dtype.name
    mac_op_type_str = "op_mac_%s_%s" % (idt_name, idt_name)
    o_mem_type_str = "mem_o_%s" % (odt_name)

    # keep in floats to remain compatible with json serialization
    n_macs, o_mem = float(n_macs), float(o_mem)
    ret = {mac_op_type_str: n_macs, o_mem_type_str: o_mem}
    return ret


def inference_cost(model, discount_sparsity=True, cost_breakdown=False):
    "Ensure all nodes have unique names prior to calling this analysis pass."

    ret, node_costs, nodes_per_optype = {}, {}, {}
    zero_cost_ops = [
        "MaxPool",
        "AveragePool",
        "Quant",
        "QuantizeLinear",
        "DequantizeLinear",
        "Reshape",
        "Concat",
        "Transpose",
        "Div",
        "Mul",
        "Add",
        "Sub",
        "BatchNormalization",
        "Relu",
        "Elu",
        "Selu",
        "Sigmoid",
        "Identity",
        "Flatten",
        "Pad",
        "Clip",
        "Trunc",
    ]
    unsupported_ops = set()
    inference_cost_fxn_map = {
        "Conv": inference_cost_conv,
        "MatMul": inference_cost_matmul,
        "Gemm": inference_cost_matmul,
        "Upsample": inference_cost_upsample,
    }
    for node in model.graph.node:
        if node.op_type in inference_cost_fxn_map.keys():
            node_cost = inference_cost_fxn_map[node.op_type](model, node, discount_sparsity)
            node_costs[node.name] = node_cost
            if node.op_type not in nodes_per_optype.keys():
                new_optype = {}
                new_optype[node.name] = node_cost
                nodes_per_optype[node.op_type] = new_optype
            else:
                nodes_per_optype[node.op_type][node.name] = node_cost
        elif node.op_type in zero_cost_ops:
            continue
        else:
            unsupported_ops.add(node.op_type)
    total = aggregate_dict_keys(node_costs)
    total["unsupported"] = unsupported_ops
    total["discount_sparsity"] = discount_sparsity
    ret["total_cost"] = total
    if cost_breakdown:
        optype_cost = {}
        for optype, resources in nodes_per_optype.items():
            optype_cost[optype] = aggregate_dict_keys(resources)
        ret["optype_cost"] = optype_cost
        ret["node_cost"] = node_costs
    return ret



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import math
import numpy as np
import onnx
import onnx.numpy_helper
from typing import Tuple

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import Transformation
from qonnx.util.basic import get_by_name


def extract_elem_type(elem_type: int, clip_range=None) -> Tuple[int, int, bool]:
    """
    Return Quant attribute specification based on element type and (optional)
    clipping range.
    Returns: (bitwidth, signed, is_narrow_qnt)
    """
    is_narrow = False
    # pylint: disable=no-member
    elem_map = {
        onnx.TensorProto.INT8: (8, 1),
        onnx.TensorProto.INT16: (16, 1),
        onnx.TensorProto.INT32: (32, 1),
        onnx.TensorProto.INT64: (64, 1),
        onnx.TensorProto.UINT8: (8, 0),
        onnx.TensorProto.UINT16: (16, 0),
        onnx.TensorProto.UINT32: (32, 0),
        onnx.TensorProto.UINT64: (64, 0),
    }
    if elem_type not in elem_map:
        raise ValueError("Unsupported element type: " + str(elem_type))
    (bw, is_signed) = elem_map[elem_type]
    if clip_range is not None:
        # refine bitwidth based on specified range
        (clip_min, clip_max) = clip_range
        # clip range elems must be scalars
        assert clip_min.ndim == 0 and clip_max.ndim == 0
        clip_min = clip_min.item()
        clip_max = clip_max.item()
        is_narrow = clip_min == -clip_max
        assert clip_min <= 0
        assert clip_max > 0
        n_repr = clip_max - clip_min + 1
        bw = int(math.ceil(math.log2(n_repr)))
    return (bw, is_signed, is_narrow)


# originally contributed by Keshav Gurushankar (@kgurushankar)
class QCDQToQuant(Transformation):
    """
    Fuse a chain of nodes, specifically QuantizeLinear+DequantizeLinear back
    into QONNX Quant node.
    This transform finds chains of QuantizeLinear followed by DequantizeLinear
    during the quantization process into a QONNX Quant node. If a Clip node is
    found between the QuantizeLinear+DequantizeLinear, this will be taken into
    account for the Quant bitwidth calculation.
    Input
    -----
    A model potentially quantized with QuantizeLinear, (optional) Clip and
    DequantizeLinear nodes.
    Output
    ------
    A model with QuantizeLinear, Clip and DequantizeLinear nodes re-fused back into QONNX
    Quant nodes.
    """

    def __init__(self) -> None:
        super().__init__()

    def apply(self, model: ModelWrapper) -> Tuple[ModelWrapper, bool]:
        graph = model.graph
        run_again = False
        nodes_to_remove = []
        for node in graph.node:
            if node.op_type == "DequantizeLinear":
                narrow = False
                dq_node = node
                dequant_node_index = model.get_node_index(dq_node)
                dq_inp, dq_scale, dq_zeropt = dq_node.input
                quant_candidates = model.find_direct_predecessors(dq_node)
                dq_init = model.get_initializer(dq_inp)
                dq_scale_v = model.get_initializer(dq_scale)
                dq_zeropt_v = model.get_initializer(dq_zeropt)
                if quant_candidates is None and dq_init is None:
                    continue
                if any([x is None for x in [dq_scale_v, dq_zeropt_v]]):
                    # unknown scale/zeropt for DQ, cannot continue
                    # TODO handle zeropt default according to spec
                    continue
                elif quant_candidates is None and dq_init is not None:
                    # standalone DequantizeLinear node with initializer
                    # (e.g. for weight quantization). we can treat this
                    # as if the QuantizeLinear node has already been
                    # constant-folded.
                    # read quantized weight dtype for standalone deqnt
                    q_vi = model.get_tensor_valueinfo(dq_inp)
                    (bitwidth, signed, narrow) = extract_elem_type(q_vi.type.tensor_type.elem_type)
                    # overwrite DQ initializer with scaled version
                    scaled_qnt_t = (dq_init - dq_zeropt_v) * dq_scale_v
                    scaled_qnt_t = scaled_qnt_t.astype(np.float32)
                    model.set_initializer(dq_inp, scaled_qnt_t)
                    q_inp = dq_inp
                    final_out = dq_node.output[0]
                    scale_factor, zeropt = dq_scale, dq_zeropt
                    nodes_to_remove.append(dq_node)
                elif quant_candidates[0].op_type in ["QuantizeLinear", "Clip"]:
                    clip_range = None
                    if quant_candidates[0].op_type == "Clip":
                        clip_node = quant_candidates[0]
                        clip_min = model.get_initializer(clip_node.input[1])
                        clip_max = model.get_initializer(clip_node.input[2])
                        if clip_min is None or clip_max is None:
                            # non-constant bounds for Clip, cannot convert
                            continue
                        clip_range = (clip_min, clip_max)
                        # keep following the producer chain
                        quant_candidates = model.find_direct_predecessors(clip_node)
                        if quant_candidates is None or quant_candidates[0].op_type != "QuantizeLinear":
                            # unexpected pattern, cannot convert
                            continue
                        # all good, mark Clip for removal and continue processing QuantizeLinear node
                        nodes_to_remove.append(clip_node)
                    quant_candidate = quant_candidates[0]
                    q_inp, q_scale, q_zeropt = quant_candidate.input
                    # check that zeropt/scale tensors are the same
                    q_scale_v = model.get_initializer(q_scale)
                    q_zeropt_v = model.get_initializer(q_zeropt)
                    if any([x is None for x in [q_scale_v, q_zeropt_v]]):
                        # TODO handle zeropt default
                        continue
                    qdq_v_match = (q_scale_v == dq_scale_v).all() and (q_zeropt_v == dq_zeropt_v).all()
                    qdq_nm_match = (q_scale == dq_scale) and (q_zeropt == dq_zeropt)
                    if not (qdq_nm_match or qdq_v_match):
                        continue
                    quant_node = quant_candidate
                    final_out = dq_node.output[0]
                    nodes_to_remove.append(dq_node)
                    nodes_to_remove.append(quant_node)
                    value_info = model.get_tensor_valueinfo(quant_node.output[0])
                    (bitwidth, signed, narrow) = extract_elem_type(value_info.type.tensor_type.elem_type, clip_range)
                    scale_factor, zeropt = q_scale, q_zeropt
                else:
                    # handle all other cases, skip
                    continue
                axis = get_by_name(dq_node.attribute, "axis")
                # fix scale factor for Quant (different shape expectations wrt broadcasting)
                if not (axis is None):
                    axis_i = axis.i
                    ishape = model.get_tensor_shape(dq_inp)
                    desired_shp = [1] * len(ishape)
                    desired_shp[axis_i] = dq_scale_v.shape[0]
                    dq_scale_v = dq_scale_v.reshape(desired_shp)
                    dq_zeropt_v = dq_zeropt_v.reshape(desired_shp)
                    model.set_initializer(scale_factor, dq_scale_v)
                    model.set_initializer(zeropt, dq_zeropt_v)
                # create new Quant node for suitable cases
                new_q_node_name = "Quant_" + q_inp
                bw_tensor_name = f"{new_q_node_name}_bitwidth"
                model.set_initializer(bw_tensor_name, np.asarray(bitwidth, dtype=np.float32))
                # for the zeropoint, see if we can simplify back to scalar
                if (dq_zeropt_v == 0).all():
                    model.set_initializer(zeropt, np.asarray(0, dtype=np.float32))
                fused_node = onnx.helper.make_node(
                    "Quant",
                    inputs=[
                        q_inp,
                        scale_factor,
                        zeropt,
                        bw_tensor_name,
                    ],
                    outputs=[final_out],
                    name=new_q_node_name,
                    domain="qonnx.custom_op.general",
                    narrow=1 if narrow else 0,  # depends on clip
                    rounding_mode="ROUND",  # round-to-even
                    signed=signed,
                )
                model.graph.node.insert(dequant_node_index, fused_node)
            for node_to_remove in nodes_to_remove:
                model.graph.node.remove(node_to_remove)
                run_again = True
            if run_again:
                break
        return (model, run_again)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import clize

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.channels_last import ConvertToChannelsLastAndClean


def to_channels_last(in_file, *, make_input_channels_last=False, out_file=None):
    """Execute a set of graph transformations to convert an ONNX file to the channels last data format.
    The input file have been previously cleaned by the cleanup transformation or commandline tool.

    :param in_file: Filename for the input ONNX model
    :param make_input_channels_last: Sets if the input of the model should also be converted to the channels
        last data layout (True) or if a transpose node should be left at the beginning of the graph (False).
        Defaults to False.
    :param out_file: If set, filename for the output ONNX model. Set to in_file with _chan_last
        suffix otherwise.
    """

    # Execute transformation
    model = ModelWrapper(in_file)
    model = model.transform(ConvertToChannelsLastAndClean(make_input_channels_last=make_input_channels_last))
    if out_file is None:
        out_file = in_file.replace(".onnx", "_channels_last.onnx")
    model.save(out_file)


def main():
    clize.run(to_channels_last)


if __name__ == "__main__":
    main()



# Copyright (c) 2023, Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.transformation.base import Transformation


class ExposeIntermediateTensorsLambda(Transformation):
    def __init__(self, tensor_filter=lambda tname, model: True):
        super().__init__()
        self.tensor_filter = tensor_filter

    def apply(self, model: ModelWrapper):
        all_tensor_names = model.get_all_tensor_names()
        for tname in all_tensor_names:
            if self.tensor_filter(tname, model):
                # check whether this tensor is already in the outputs
                if tname in [x.name for x in model.graph.output]:
                    # already part of outputs, skip
                    continue
                else:
                    # append ValueInfo to outputs
                    tensor_vi = model.get_tensor_valueinfo(tname)
                    model.graph.output.append(tensor_vi)
                    # remove existing ValueInfo to avoid duplicate
                    model.graph.value_info.remove(tensor_vi)

        return (model, False)


class ExposeIntermediateTensorsPatternList(ExposeIntermediateTensorsLambda):
    def pattern_filter(self, tname, model):
        if self.dynamic_only:
            return any([(pat in tname) and (model.get_initializer(tname) is None) for pat in self.pattern_list])
        else:
            return any([(pat in tname) for pat in self.pattern_list])

    def __init__(self, pattern_list, dynamic_only=True):
        self.pattern_list = pattern_list
        self.dynamic_only = dynamic_only
        super().__init__(tensor_filter=self.pattern_filter)



# Copyright (c) 2023 Advanced Micro Devices, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of qonnx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import numpy as np
from onnx import TensorProto
from onnx import helper as oh
from warnings import warn

from qonnx.core.modelwrapper import ModelWrapper
from qonnx.custom_op.general.quant import max_int, min_int
from qonnx.custom_op.registry import getCustomOp
from qonnx.transformation.base import Transformation
from qonnx.transformation.general import MovePadAttributeToTensor, RemoveUnusedTensors


class QuantToQCDQ(Transformation):
    """Replace QONNX Quant-style quantization nodes with QuantizeLinear
    -> Clip -> DequantizeLinear (QCDQ)-style quantization nodes. The following
    restictions apply on the Quant:
    - the scale, zero-point and bitwidth inputs for Quant must be statically specified
      by an initializer
    - the bitwidth must be an integer in the range [2, 8]
    - the zero-point tensor must be zero
    - the scale must be a scalar value or 1D tensor
    - the rounding_mode attribute must be ROUND
    BipolarQuant is not (yet) supported.
    """

    def apply(self, model: ModelWrapper):
        graph = model.graph
        graph_modified = False
        nodes_to_remove = []
        for node in graph.node:
            if node.op_type == "Quant":
                # check all conditions for conversion
                inp_tname = node.input[0]
                scale_tname = node.input[1]
                zeropt_tname = node.input[2]
                bitwidth_tname = node.input[3]
                out_tname = node.output[0]
                scale_t = model.get_initializer(scale_tname)
                zeropt_t = model.get_initializer(zeropt_tname)
                bitwidth_t = model.get_initializer(bitwidth_tname)
                qnt_inst = getCustomOp(node)
                narrow = bool(qnt_inst.get_nodeattr("narrow"))
                signed = bool(qnt_inst.get_nodeattr("signed"))
                rounding_mode = qnt_inst.get_nodeattr("rounding_mode")
                last_node = node
                if zeropt_t is None or zeropt_t != 0:
                    warn("Zeropoint is undefined or nonzero, skipping " + node.name)
                    continue
                if scale_t is None or (scale_t.squeeze().ndim not in [0, 1]):
                    warn("Scale is undefined or non-0/1D, skipping " + node.name)
                    continue
                if bitwidth_tname is None or bitwidth_t.ndim != 0:
                    warn("Bitwidth is undefined or non-scalar, skipping " + node.name)
                    continue
                if bitwidth_t < 2 or bitwidth_t > 8:
                    warn("Bitwidth outside supported range [2,8], skipping " + node.name)
                    continue
                if rounding_mode != "ROUND":
                    warn("Rounding mode is not ROUND, skipping " + node.name)
                    continue
                graph_modified = True
                # QuantizeLinear wants 1D scale so squeeze
                q_scale = model.get_initializer(scale_tname)
                q_scale_new = q_scale.squeeze()
                model.set_initializer(scale_tname, q_scale_new)
                # resolve the scale factor axis
                if q_scale_new.ndim == 1:
                    qnt_axis = q_scale.shape.index(q_scale_new.shape[0])
                else:
                    qnt_axis = None
                # create new zeropoint tensor with appropriate dtype
                new_dtype = TensorProto.INT8 if signed else TensorProto.UINT8
                new_dtype_np = np.int8 if signed else np.uint8
                new_zeropt_name = model.make_new_valueinfo_name()
                new_zeropt_vi = oh.make_tensor_value_info(new_zeropt_name, new_dtype, scale_t.shape)
                new_zeropt_t = np.zeros_like(q_scale_new, dtype=new_dtype_np)
                graph.value_info.append(new_zeropt_vi)
                model.set_initializer(new_zeropt_name, new_zeropt_t)
                new_qout_name = model.make_new_valueinfo_name()
                ishape = model.get_tensor_shape(inp_tname)
                new_qout_vi = oh.make_tensor_value_info(new_qout_name, new_dtype, ishape)
                graph.value_info.append(new_qout_vi)
                n_ql = len(model.get_nodes_by_op_type("QuantizeLinear"))
                n_cl = len(model.get_nodes_by_op_type("Clip"))
                n_dql = len(model.get_nodes_by_op_type("DequantizeLinear"))
                # create the QuantizeLinear node
                new_node_qnt = oh.make_node(
                    "QuantizeLinear",
                    [inp_tname, scale_tname, new_zeropt_name],
                    [new_qout_name],
                    name="QuantizeLinear_%d" % (n_ql + 1),
                    axis=qnt_axis,
                )
                graph.node.insert(model.get_node_index(last_node) + 1, new_node_qnt)
                last_out_tname = new_qout_name
                last_node = new_node_qnt
                # determine whether clipping is needed
                # insert a Clip node if so
                bitwidth = int(bitwidth_t)
                range_min = min_int(signed, narrow, bitwidth)
                range_max = max_int(signed, narrow, bitwidth)
                if (signed and narrow) or (bitwidth < 8):
                    new_clip_oname = model.make_new_valueinfo_name()
                    new_clip_vi = oh.make_tensor_value_info(new_clip_oname, new_dtype, ishape)
                    graph.value_info.append(new_clip_vi)
                    new_clip_min = model.make_new_valueinfo_name()
                    new_clip_min_t = np.asarray(range_min, dtype=new_dtype_np)
                    # set_initializer will create the missing VIs for the
                    # clip min/max tensors here
                    model.set_initializer(new_clip_min, new_clip_min_t)
                    new_clip_max = model.make_new_valueinfo_name()
                    new_clip_max_t = np.asarray(range_max, dtype=new_dtype_np)
                    model.set_initializer(new_clip_max, new_clip_max_t)
                    new_node_clip = oh.make_node(
                        "Clip", [last_out_tname, new_clip_min, new_clip_max], [new_clip_oname], name="Clip_%d" % (n_cl + 1)
                    )
                    last_out_tname = new_clip_oname
                    graph.node.insert(model.get_node_index(last_node) + 1, new_node_clip)
                    last_node = new_node_clip
                # finally add the DequantizeLinear node
                new_node_deqnt = oh.make_node(
                    "DequantizeLinear",
                    [last_out_tname, scale_tname, new_zeropt_name],
                    [out_tname],
                    name="DequantizeLinear_%d" % (n_dql + 1),
                    axis=qnt_axis,
                )
                graph.node.insert(model.get_node_index(last_node) + 1, new_node_deqnt)
                last_node = new_node_deqnt
                nodes_to_remove.append(node)
        for node_to_remove in nodes_to_remove:
            graph.node.remove(node_to_remove)
        if graph_modified:
            # remove old tensors that are no longer used for bitwidth, zeropt etc
            model = model.transform(RemoveUnusedTensors())
            # ensure opset version is new enough for the Clip nodes we inserted
            clip_min_opset = 13
            if model.model.opset_import[0].version < clip_min_opset:
                warn(
                    "QCDQ Clip requires ONNX opset >= %d but found %d"
                    % (clip_min_opset, model.model.opset_import[0].version)
                )
                warn("Forcing opset version %s upgrade to ensure valid ONNX" % clip_min_opset)
                model.model.opset_import[0].version = clip_min_opset
                # ensure new Pad node requirements are respected
                model = model.transform(MovePadAttributeToTensor())

        return (model, graph_modified)



# Copyright (c) 2020, Xilinx
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of QONNX nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json

from qonnx.custom_op.registry import getCustomOp


def extract_model_config_to_json(model, json_filename, attr_names_to_extract):
    """Create a json file with layer name -> attribute mappings extracted from the
    model. The created json file can be later applied on a model with
    qonnx.transform.general.ApplyConfig."""

    cfg = dict()
    cfg["Defaults"] = dict()
    for n in model.graph.node:
        oi = getCustomOp(n)
        layer_dict = dict()
        for attr in attr_names_to_extract:
            try:
                layer_dict[attr] = oi.get_nodeattr(attr)
            except AttributeError:
                pass
        if len(layer_dict) > 0:
            cfg[n.name] = layer_dict
    with open(json_filename, "w") as f:
        json.dump(cfg, f, indent=2)



# Copyright (c) 2020 Xilinx, Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of Xilinx nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

"""
Guide to writing QONNX transformations
---------------------------------------

* Your transformation must inherit the Transformation abstract base class.
* Your transformation's apply function should take in a ModelWrapper, and return
  a tuple with (transformed_model: ModelWrapper, model_was_changed: Bool)
* The transformations are meant to be applied using the .transform function
  in ModelWrapper. This makes a deep copy of the input model by default, so
  you don't have to.
* model_was_changed indicates whether your transformation made any changes to
  the model. If you know your transformation needs to be called only once and
  repeated calls have no further effect, you can return False even if the model
  was changed.
* You MUST return model_was_changed=False at some point when your transformation
  is called multiple times, otherwise apply_repeated() will loop infinitely.
* If you cannot guarantee that the transformation will reach a fixed point,
  you must declare this, return model_was_changed = False and let the user
  manually re-apply the transform.
"""

import copy
import multiprocessing as mp
from abc import ABC, abstractmethod

from qonnx.util.basic import get_num_default_workers


class Transformation(ABC):
    """Transformation class all transformations are based on. Contains only
    abstract method apply() every transformation has to fill."""

    def __init__(self):
        super().__init__()

    @abstractmethod
    def apply(self, model):
        pass


class NodeLocalTransformation(Transformation):
    """
    Parent class for transformations, which can be executed locally to one node
    by accessing and modifying the attributes of only that node.
    This class can then automatically parallelize the transformation.
    Transformations sublcassing NodeLocalTransformation must implement the
    abstract method applyNodeLocal(). A read-only copy of the model is available
    as a member variable ref_input_model, but any modifications there will be
    disregarded.

    To control the degree of parallelization, specify the num_workers argument
    in the constructor, using one of the following values:
    * None: use NUM_DEFAULT_WORKERS environment variable
    * 0: use all available CPU cores
    * (any other int>0): set number of parallel workers
    """

    def __init__(self, num_workers=None):
        super().__init__()
        if num_workers is None:
            self._num_workers = get_num_default_workers()
        else:
            self._num_workers = num_workers
        assert self._num_workers >= 0, "Number of workers must be nonnegative."
        if self._num_workers == 0:
            self._num_workers = mp.cpu_count()

    @abstractmethod
    def applyNodeLocal(self, node):
        pass

    def apply(self, model):
        # make a detached copy of the input model that applyNodeLocal
        # can use for read-only access
        self.ref_input_model = copy.deepcopy(model)
        # Remove old nodes from the current model
        old_nodes = []
        for i in range(len(model.graph.node)):
            old_nodes.append(model.graph.node.pop())

        # Execute transformation in parallel
        if self._num_workers > 1:
            with mp.Pool(self._num_workers) as p:
                new_nodes_and_bool = p.map(self.applyNodeLocal, old_nodes, chunksize=1)
        # execute without mp.Pool in case of 1 worker to simplify debugging
        else:
            new_nodes_and_bool = [self.applyNodeLocal(node) for node in old_nodes]

        # extract nodes and check if the transformation needs to run again
        # Note: .pop() had initially reversed the node order
        run_again = False
        for node, run in reversed(new_nodes_and_bool):
            # Reattach new nodes to old model
            model.graph.node.append(node)
            if run is True:
                run_again = True

        return (model, run_again)
